#include "udp.h"

#include "eth_phy.h"
#include "hardware_profile.h"

/*****************************************************************************
 * Resources for Ethernet Controller Descriptors
 * Can modify as needed.         
 *****************************************************************************/
#define EMAC_TX_DESCRIPTORS		10		// number of the TX descriptors to be created
#define EMAC_RX_DESCRIPTORS		1		// number of the RX descriptors to be created 

#define	EMAC_RX_BUFF_SIZE		1536	// size of a RX buffer. should be multiple of 16
										// this is the minimum size of all receive buffers
										// that can be passed by the stack to the ETHC
										// If the packets are large, they will take multiple RX buffers and the overhead will be higher

#define EMAC_OPEN_LINK      (ETH_OPEN_AUTO|ETH_OPEN_10|ETH_OPEN_100|ETH_OPEN_HDUPLEX|ETH_OPEN_FDUPLEX|ETH_OPEN_MDIX_AUTO)
										// flags to be used for the Ethernet link
										// adjust to your needs. ETH_OPEN_AUTO (auto negotiation) strongly recommended.
										// Note: if Auto MDIX is not enabled, ETH_OPEN_MDIX_SWAP or ETH_OPEN_MDIX_NORM could be selected
/** LOCAL DATA ****************************************************/

// default opening modes
//
static eEthOpenFlags    _stackOFlags=EMAC_OPEN_LINK;      // default opening flags

#ifdef PHY_CONFIG_ALTERNATE
static eEthPhyCfgFlags  _stackCfgFlags=ETH_PHY_CFG_RMII|ETH_PHY_CFG_ALTERNATE;  
#else
static eEthPhyCfgFlags  _stackCfgFlags=ETH_PHY_CFG_RMII|ETH_PHY_CFG_DEFAULT;  
#endif
// default PHY configuration


static eEthRxFilters        _stackRxFilters=(ETH_FILT_CRC_ERR_REJECT|ETH_FILT_RUNT_REJECT|ETH_FILT_ME_UCAST_ACCEPT|ETH_FILT_MCAST_ACCEPT|ETH_FILT_BCAST_ACCEPT);
// default RX filtering


static sBuffDcpt	rxPktDcpt[EMAC_RX_DESCRIPTORS+2];		// rx buffer descriptor
static u8*			rxBuffers[EMAC_RX_DESCRIPTORS] = {0};	// rx buffers for incoming data



// timing and link status maintenance
eEthLinkStat     _linkPrev;                                 // last value of the link status
static int              _linkPresent;                       // if connection to the PHY properly detected
static int              _linkNegotiation;                   // if an auto-negotiation is in effect

//prote type def
static int      _LinkReconfigure(int waitComplete);                     // link reconfiguration
DWORD TCPIPAddrStrToDword(const char* szDot);


IP_ADDR     SysIPAddr;		// big endian
IP_ADDR 	SysIPMask;		// big-endian
IP_ADDR 	SysIPGateway;	// big-endian
MAC_ADDR    SysMACAddr;		// big-endian

IP_ADDR		TargetIPAddr;	// big-endian
MAC_ADDR    TargetMACAddr;	// big-endian

/****************************************************************************
 * Function:        MACClose
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *                
 * Side Effects:    None
 *
 * Overview:        This function performs a clean-up of the resources allocated by the MAC
 *                  It also disables Eth controller, without changing the set-up.
 *                  Any Eth transmit or receive operation i saborted.
 *
 * Note:            None 
 *****************************************************************************/
void MACClose(void)
{
        int ix;
        
        EthClose(0);

        EthDescriptorsCleanUp(ETH_DCPT_TYPE_ALL, 0);
}


/****************************************************************************
 * Function:        MACInit
 *
 * PreCondition:    None
 *
 * Input:           waitComplete - wait for auto-negotiation to complete
 *
 * Output:          None
 *                
 * Side Effects:    None
 *
 * Overview:        This function initializes the Eth controller, the MAC and the PHY. It should be called to be able to schedule
 *                  any Eth transmit or receive operation.
 *
 * Note:           The flags used for EthOpen() call are fixed for now. This will change in the future! 
 *****************************************************************************/
int MACInit(int waitComplete)
{
    int     ix;
	sBuffDcpt*	pRxPkt;
    eEthRes     ethRes, phyInitRes;
    int     success=0;

    _linkNegotiation=_linkPresent=0;
    _linkPrev=ETH_LINK_ST_DOWN;
    
    while(1)
    {
        eEthLinkStat    linkStat;
        eEthOpenFlags   linkFlags;
        eMacPauseType   pauseType;
        

        pauseType=(_stackOFlags&ETH_OPEN_FDUPLEX)?MAC_PAUSE_CPBL_MASK:MAC_PAUSE_TYPE_NONE;
        
        // start the initialization sequence    
        EthInit();

        phyInitRes=EthPhyInit(_stackOFlags, _stackCfgFlags, &linkFlags);

        if(phyInitRes!=ETH_RES_OK)
        {
	        assert(0);
        }

        // let the auto-negotiation (if any) take place
        // continue the initialization
        EthRxFiltersClr(ETH_FILT_ALL_FILTERS);
        EthRxFiltersSet(_stackRxFilters);

        
        EthMACSetAddress(SysMACAddr.v);
                
        if(EthDescriptorsAdd(EMAC_TX_DESCRIPTORS, ETH_DCPT_TYPE_TX, 0) != EMAC_TX_DESCRIPTORS)
        {
	        assert(0);
        }

        if(EthDescriptorsAdd(EMAC_RX_DESCRIPTORS, ETH_DCPT_TYPE_RX, 0) != EMAC_RX_DESCRIPTORS)
        {
	        assert(0);
        }

        EthRxSetBufferSize(EMAC_RX_BUFF_SIZE);


		memset(rxBuffers, 0, sizeof(rxBuffers));	// clear buffer pointers
		memset(rxPktDcpt, 0, sizeof(rxPktDcpt));	// clear packet descriptors

		for(ix=0, pRxPkt=rxPktDcpt; ix<EMAC_RX_DESCRIPTORS; ix++){
			pRxPkt->next=pRxPkt+1;
			pRxPkt=pRxPkt->next;
		}
		rxPktDcpt[ix-1].next=0;	// end properly

		// allocate memory for the RX buffers
		for(ix=0; ix<EMAC_RX_DESCRIPTORS; ix++)
		{
			if(!(rxBuffers[ix]=(unsigned char*)malloc(EMAC_RX_BUFF_SIZE)))
			{
				assert(0);
			}
		}

		// set these buffers as permanent receive buffers
		ethRes=EthRxBuffersAppend(rxBuffers, EMAC_RX_DESCRIPTORS, BUFF_FLAG_RX_STICKY);
		if(ethRes!=ETH_RES_OK)
		{
			break;
		}

		// continue the MAC set-up based on the (possible) PHY negotiation
		_linkPresent=1;     // show we have valid connection to the PHY
		if(_stackOFlags&ETH_OPEN_AUTO)
		{   // have to wait the negotiation to be done
			_linkNegotiation=1;     // PHY performing negotiation if unplugged/plugged       
			linkStat=_LinkReconfigure(waitComplete)?ETH_LINK_ST_UP:ETH_LINK_ST_DOWN;    // if negotiation not done yet we need to try it next time
		}
		else
		{   // no need of negotiation results; just update the MAC 
			EthMACOpen(linkFlags, pauseType);
			linkStat=EthPhyGetLinkStatus(0);
		}

		_linkPrev=linkStat;

		success=1;
		break;
	}

	if(!success){
		MACClose();
	}

	return success;

}


/****************************************************************************
 * Function:        MACIsLinked
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          TRUE if link is up
 *                  FALSE otherwise
 *
 * Side Effects:    None
 *
 * Overview:        This function checks the link status
 *
 * Note:            None 
 *****************************************************************************/
BOOL MACIsLinked(void)
{
    return (_linkPrev&ETH_LINK_ST_UP)!=0;
}



/****************************************************************************
 * Function:        MACCheckLink
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          TRUE if link is up
 *                  FALSE otherwise
 *
 * Side Effects:    None
 *
 * Overview:        This function periodically checks the link status
 *                  performing the MAC reconfiguration if the link went up
 *                  after being down.
 *
 * Note:            None 
 *****************************************************************************/
BOOL MACCheckLink(void)
{
    while(_linkPresent)
    {
        eEthLinkStat    linkCurr;

        linkCurr=EthPhyGetLinkStatus(0);    // read current PHY status
        if(_linkNegotiation)
        {   // the auto-negotiation turned on
            if((linkCurr&ETH_LINK_ST_UP) && !(_linkPrev&ETH_LINK_ST_UP))
            {   // we're up after being done. do renegotiate!
                linkCurr=_LinkReconfigure(0)?ETH_LINK_ST_UP:ETH_LINK_ST_DOWN;    // if negotiation not done yet we need to try it next time
            }
            // else link went/still down; nothing to do yet
        }
        _linkPrev=linkCurr;
    
        break;
    }

    return (_linkPrev&ETH_LINK_ST_UP)!=0;

}



/*********************************************************************
* Function:        int  _LinkReconfigure(int waitComplete)
 *
 * PreCondition:    None
 * 
 * Input:           waitComplete    -  wait for negotiation to be done
 * 
 * Output:          TRUE if negotiation succeeded and MAC was updated
 *                  FALSE otherwise
 * 
 * Side Effects:    None
 * 
 * Overview:        Performs re-configuration after auto-negotiation performed.
 * 
 * Note:            None
 ********************************************************************/
static int _LinkReconfigure(int waitComplete)
{

    eEthOpenFlags   linkFlags;
    eEthLinkStat    linkStat;
    eMacPauseType   pauseType;
    int             success=0;

    if(EthPhyNegotiationComplete(waitComplete)==ETH_RES_OK)
    {    // negotiation complete
        linkStat=EthPhyGetNegotiationResult(&linkFlags, &pauseType);
        if(linkStat&ETH_LINK_ST_UP)
        {   // negotiation succeeded; properly update the MAC
            EthMIIMInit(sys_freq, EthPhyMIIMClock(), linkFlags, 1);
            EthMACOpen(linkFlags, pauseType);
            success=1;
        }
    }

    return success;
}
/****************************************************************************
 * Function:        TCPIPAddrStrToDword
 *
 * PreCondition:    None
 *
 * Input:           szDot - an IP address in standard dot notation 
 *
 * Output:          The IP address as an BE dword.
 *
 * Side Effects:    None
 *
 * Overview:        Converts a dot-address string to big-endian dword.
 *
 * Note:            None
 * 
 * Example:        DWORD myAddr=TCPIPAddrStrToDword("192.168.2.105");         
 ******************************************************************************/
DWORD TCPIPAddrStrToDword(const char* szDot)
{
	DWORD_VAL dw;
	const char* pip;
	char c;

    int index = 0;
	BYTE acc = 0;
	dw.Val = 0;
	pip = szDot;
	
	while( (c = *pip++) ) 
	{
        if( c == '.' )
		{
            dw.v[index++] = acc;
            acc = 0;
        }
		else if( c == ' ' )
			continue;
		else 
            acc = acc*10 + (c - '0');
    }

    dw.v[3] = acc;
    return dw.Val;

}

static u8	*arp_icmp_buf;

struct ARP{
	u16	hwtype;
	u16	protocol;
	u8	hwlen;
	u8	protolen;
	u16	opcode;
	u8	s_hwaddr[6];
	u16	s_ipaddr[2];
	u8	t_hwaddr[6];
	u16	t_ipaddr[2];
};

struct ICMP{
	u8	type;
	u8	code;
	u16	chksum;
	u16	ident;
	u16	seq;
};

#define UDP_PACKET_SIZE	(188*256)	//6*42+4=256
#define UDP_PORT		(1234)

sBuffDcpt	txh_dcpt[8];
sBuffDcpt	txd_dcpt[8];
u32			udp_idt_val;

_tx_stat	tx_buf_stat[8];	//TX_DONE, TX_STORED, TX_SENDING

void udp_init(){
	u32 i, len;
	unsigned short* pS=(unsigned short*)(&SysMACAddr.v);
	
	// read factory writed address
	*pS++=EMACxSA2;
	*pS++=EMACxSA1;
	*pS=EMACxSA0;
	
	// Initialize default IP and MAC address
	SysIPAddr.Val =TCPIPAddrStrToDword(DEFAULT_IP_ADDR);	
	SysIPMask.Val =TCPIPAddrStrToDword(DEFAULT_IP_MASK);	
	SysIPGateway.Val=TCPIPAddrStrToDword(DEFAULT_IP_GATEWAY);
	
	TargetIPAddr.Val =TCPIPAddrStrToDword(TARGET_IP);
	memset(&TargetMACAddr, 0, sizeof(MAC_ADDR));

	MACInit(1);

	if(!(arp_icmp_buf = (u8 *)malloc(1514))){
		assert(0);
	}
	
	len = sizeof(struct ETH) + sizeof(struct IPH) + sizeof(struct UDP);
	for(i=0; i<8; i++){
		//ETH+IP+UDP wb_̃AP[g
		if(!(txh_dcpt[i].pBuff = (void *)malloc(len))){
			assert(0);
		}
		txh_dcpt[i].nBytes = len;
		txh_dcpt[i].next = &(txd_dcpt[i]);
		
		//UDPf[^{
		txd_dcpt[i].next = NULL;
		tx_buf_stat[i] = TX_DONE;
	}
}

u16 calc_chksum(u16 *buf, int bufsz){
	u32 sum = 0;

	while (bufsz > 1) {
		sum += *buf;
		buf++;
		bufsz -= 2;
	}
	if (bufsz == 1) {
		sum += *(u8 *)buf;
	}
	sum = (sum & 0xffff) + (sum >> 16);
	sum = (sum & 0xffff) + (sum >> 16);
	return ~sum;
}

static u16 ident = 0;

void ip_build(u8 *p, u32 proto, u8 *hwaddr, u16 *ipaddr, u32 len, u32 flags_ofs){
	struct ETH *eth;
	struct IPH *iph;

	eth=(void *)p;
	memcpy(eth->t_hwaddr, hwaddr, 6);
	memcpy(eth->s_hwaddr,&SysMACAddr, 6);
	eth->type=0x0008;
	iph=(void *)eth + sizeof(struct ETH);
	iph->version=0x45;
	iph->service=0;
	iph->len=((len-sizeof(struct ETH)) >> 8) + ((len-sizeof(struct ETH)) << 8);
	iph->ident=(ident >> 8) + (ident << 8);
	iph->frags=(flags_ofs >> 8)+(flags_ofs << 8);
	iph->ttl=0x08;
	iph->protocol=proto;
	iph->chksum=0;
	iph->src_addr[0] = SysIPAddr.w[0];
	iph->src_addr[1] = SysIPAddr.w[1];
	iph->dest_addr[0] = ipaddr[0];
	iph->dest_addr[1] = ipaddr[1];
	iph->chksum=calc_chksum((u16 *)iph, sizeof(struct IPH));
	ident = (ident + 1) & 0x7fff;
}

void udp_build(u8 *p, u32 dest_port, u32 src_port, u32 len){
	struct UDP *udp;
	udp = (void *)p  + sizeof(struct ETH) + sizeof(struct IPH);
	udp->src_port = (src_port >> 8) + (src_port << 8);
	udp->dest_port = (dest_port >> 8) + (dest_port << 8);
	len += sizeof(struct UDP);
	udp->len = (len >> 8) + (len << 8);
	udp->chksum = 0;
}

void mk_udp_header(){
	struct	IPH	*iph;
	u32 		i;

	printf("Target[%d.%d.%d.%d]\n", TargetIPAddr.v[0], TargetIPAddr.v[1], TargetIPAddr.v[2], TargetIPAddr.v[3]);
	udp_idt_val = (ReadCoreTimer() & 0xfffc) | 0x8000;
	for(i=0; i<8; i++){
		ip_build(txh_dcpt[i].pBuff, 0x11, TargetMACAddr.v, TargetIPAddr.w,
			 sizeof(struct ETH) + sizeof(struct IPH) + sizeof(struct UDP) + 1472, 4);
		udp_build(txh_dcpt[i].pBuff, UDP_PORT, UDP_PORT, UDP_PACKET_SIZE);
	}
}


static u32			arp_icmp_buf_used = 0;

void udp_rcv_proc(){
	u32					i;
	int					buffsPerRxPkt;
	const sRxPktStat*	pRxPktStat;
	eEthRes				res;
	u32					len;
	struct ETH 			*eth, *t_eth;
	struct ARP 			*arp, *t_arp;
	struct IPH			*iph;
	struct ICMP			*icmp;

	if(arp_icmp_buf_used){
		res=EthTxAckBuffer(arp_icmp_buf, 0);
		arp_icmp_buf_used = 0;
		if(res != ETH_RES_OK){
			assert(0);
		}
	}
	
    res=EthRxGetPacket(rxPktDcpt, &buffsPerRxPkt, &pRxPktStat);
    if(res==ETH_RES_OK){   // available packet; minimum check
	    len = pRxPktStat->rxBytes;
		eth = rxPktDcpt[0].pBuff;
	    if(pRxPktStat->bcast){
			if(eth->type == 0x0608){//ARPv
				arp = (void *)eth + 14;
				if(arp->opcode == 0x0100){
					if((arp->t_ipaddr[0] == SysIPAddr.w[0]) && (arp->t_ipaddr[1] == SysIPAddr.w[1])){
						memcpy(arp_icmp_buf, eth, len);
						t_eth = (void *)arp_icmp_buf;
						t_arp = (void *)arp_icmp_buf + 14;
						memcpy(t_arp->t_hwaddr, arp->s_hwaddr, 10);
						memcpy(t_arp->s_hwaddr, &SysMACAddr, 6);
						t_arp->s_ipaddr[0] = SysIPAddr.w[0];
						t_arp->s_ipaddr[1] = SysIPAddr.w[1];
						t_arp->opcode=0x0200;
						memcpy(t_eth->t_hwaddr, eth->s_hwaddr, 6);
						memcpy(t_eth->s_hwaddr, &SysMACAddr, 6);
						res=EthTxSendBuffer(arp_icmp_buf, len);
						if(res != ETH_RES_OK){
							assert(0);
						}
						else{
							arp_icmp_buf_used = 1;
						}
						TargetIPAddr.w[0] = arp->s_ipaddr[0];
						TargetIPAddr.w[1] = arp->s_ipaddr[1];
						memcpy(&TargetMACAddr, arp->s_hwaddr, 6);
						mk_udp_header();
						printf("ARP(req)\n");
					}
				}
			}
		}
		else if(eth->type == 0x0608){
			arp = (void *)eth + 14;
			if(arp->opcode == 0x0200){//ARP
				if((arp->s_ipaddr[0] == TargetIPAddr.w[0]) && (arp->s_ipaddr[1] == TargetIPAddr.w[1])){
					memcpy(&TargetMACAddr, arp->s_hwaddr, 6);
					mk_udp_header();
					printf("ARP(rpl)\n");
				}
			}
		}
		else if(eth->type == 0x0008){
			iph = (void *)eth + 14;
			if((iph->version & 0xf0) == 0x40){//Version 4
				if((iph->dest_addr[0] == SysIPAddr.w[0]) && (iph->dest_addr[1] == SysIPAddr.w[1])){//IP
					if(iph->protocol==1){//ICMP
						icmp=(void *)((u32)iph+(iph->version & 0x0f)*4);
						if(icmp->type==8){//Echo Request
							icmp->type=0;
							icmp->code=0;
							icmp->chksum=0;
							icmp->chksum=calc_chksum((u16 *)icmp, len-14-(iph->version & 0x0f)*4);
							memcpy(arp_icmp_buf+sizeof(struct ETH)+sizeof(struct IPH), icmp, len-14-(iph->version & 0x0f)*4);
							ip_build(arp_icmp_buf, 1, eth->s_hwaddr, iph->src_addr, len, 0);
							res=EthTxSendBuffer(arp_icmp_buf, len);
							if(res != ETH_RES_OK){
								assert(0);
							}
							else{
								arp_icmp_buf_used = 1;
							}
							printf("Echo\n");
						}
					}
				}
			}
		}

		EthRxAckPacket(rxPktDcpt, NULL);
	}
}

void arp_request(u16 *ipaddr){
	eEthRes		res;
	struct ETH	*eth;
	struct ARP 	*arp;
	
	if(arp_icmp_buf_used){
		res=EthTxAckBuffer(arp_icmp_buf, 0);
		arp_icmp_buf_used = 0;
		if(res != ETH_RES_OK){
			assert(0);
		}
	}
	
	eth = (void *)arp_icmp_buf;
	arp = (void *)eth + sizeof(struct ETH);
	
	memset(eth->t_hwaddr, 0xff, 6);
	memcpy(eth->s_hwaddr, &SysMACAddr, 6);
	eth->type = 0x0608;
	arp->hwtype = 0x0100;
	arp->protocol = 0x0008;
	arp->hwlen = 0x06;
	arp->protolen = 0x04;
	arp->opcode = 0x0100;
	memcpy(arp->s_hwaddr, &SysMACAddr, 6);
	arp->s_ipaddr[0] = SysIPAddr.w[0];
	arp->s_ipaddr[1] = SysIPAddr.w[1];
	memset(arp->t_hwaddr, 0x00, 6);
	arp->t_ipaddr[0] = ipaddr[0];
	arp->t_ipaddr[1] = ipaddr[1];

	res=EthTxSendBuffer(arp_icmp_buf, sizeof(struct ETH) + sizeof(struct ARP));
	if(res != ETH_RES_OK){
		assert(0);
	}
	else{
		arp_icmp_buf_used = 1;
	}
}

void udp_trans(){
	static u32	cnt = 0;
	static u32	_cnt = 0;
	eEthRes		res;
	
	if(tx_buf_stat[_cnt] == TX_SENDING){
		res = EthTxGetBufferStatus(txh_dcpt[_cnt].pBuff, NULL);
		if(res != ETH_RES_PACKET_QUEUED ){
			tx_buf_stat[_cnt] = TX_DONE;
			EthTxAckBuffer(txh_dcpt[_cnt].pBuff, NULL);
			if(++_cnt == 8) _cnt = 0;
		}	
	}

	if(tx_buf_stat[cnt] == TX_STORED){
		res=EthTxSendPacket(&(txh_dcpt[cnt]));
		if(res != ETH_RES_OK){
			assert(0);
		}
		tx_buf_stat[cnt] = TX_SENDING;
		if(++cnt == 8) cnt = 0;
	}
}
